import 'dart:collection';
import 'dart:io' as io;

import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:flutter_localizations/flutter_localizations.dart';
import 'package:flutterx_application/src/app_route.dart';
import 'package:flutterx_application/src/localization/labels.dart';
import 'package:flutterx_live_data/flutterx_live_data.dart';
import 'package:flutterx_utils/flutterx_utils.dart';

part 'app/application_binding.dart';
part 'app/application_impl.dart';

/// Unique representation of your application
///
/// To use this class just call [runMaterialApp] in your main() method instead of [runApp]
///
/// [of] use this to get [Application] instance from everywhere in your app
/// [isInputMethodVisible] [LiveData] that represents true if input method is visible
/// [frameCallback] [LiveEvent] you can use to listen [WidgetsBinding.addPersistentFrameCallback]
/// [supportedLocales] list of [Locale] supported by this application
/// [defaultLocale] is default locale resolved from supported locales
/// [enforcedLocale] can be used to force Application locale regardless of system settings, or null for default
/// [restart] (soft) recreate widget tree containing the whole application. This will not kill application process
/// [exit] close application. (this can kill application process)
/// [addLifecycleObserver] add [LifecycleObserver] and immediately dispatch current [AppLifecycleState] to it
/// [removeLifecycleObserver] remove [LifecycleObserver]
abstract class Application {
  static Application of(BuildContext context) =>
      requireNotNull(context.findRootAncestorStateOfType<_ApplicationImplState>());

  LiveData<bool> get isInputMethodVisible;

  LiveEvent<Duration> get frameCallback;

  List<Locale> get supportedLocales;

  Locale get defaultLocale;

  Locale? enforcedLocale;

  void restart();

  void exit();

  void addLifecycleObserver(LifecycleObserver observer);

  void removeLifecycleObserver(LifecycleObserver observer);

  /// Compute default [SystemUiOverlayStyle] from [context]
  static SystemUiOverlayStyle defaultSystemUiStyle(BuildContext context) {
    final theme = Theme.of(context);
    final brightness = theme.appBarTheme.brightness;
    final navBarColor = theme.scaffoldBackgroundColor.darken(.05);
    return SystemUiOverlayStyle(
        systemNavigationBarColor: navBarColor,
        systemNavigationBarIconBrightness: navBarColor.brightness.negative,
        statusBarIconBrightness: brightness?.negative,
        statusBarBrightness: brightness);
  }
}

/// [Application] observes [AppLifecycleState] through [WidgetsBindingObserver] and notify changes to
/// [LifecycleObserver] registered by calling [Application.addLifecycleObserver].
typedef LifecycleObserver = void Function(AppLifecycleState);

/// Create [SystemUiOverlayStyle] from context.
/// This is used by [AppRoute], default value is [Application.defaultSystemUiStyle]
typedef SystemUiStyleBuilder = SystemUiOverlayStyle Function(BuildContext context);

/// Factory of [MaterialApp] used as parameter of [runMaterialApp]
///
/// Every parameter passed by this closure must be used as parameter of the returned [MaterialApp]
typedef MaterialAppFactory = MaterialApp Function(
    GlobalKey key,
    String initialRoute,
    RouteFactory onGenerateRoute,
    String title,
    Locale? locale,
    Iterable<LocalizationsDelegate> localizationsDelegates,
    Iterable<Locale> supportedLocales);

/// [runApp] wrapping the widget tree inside [Application] implementation
///
/// [name] of the application
/// [initialRoute] the first route of the application
/// [labels] set of supported [LabelInterface] used inside the application
/// [onLabel] callback called when application [Locale] changes passing the corresponding [LabelInterface].
///           Use this callback to update the labels global variable
/// [materialAppFactory] check [MaterialAppFactory].
///                      Default value uses the default theme so you may want to customize it
void runMaterialApp(
        {required String name,
        required AppRoute initialRoute,
        Set<LabelInterface>? labels,
        ValueSetter<LabelInterface>? onLabel,
        MaterialAppFactory materialAppFactory = _defaultMaterialAppFactory}) =>
    runApp(_ApplicationImpl(
        name: name,
        initialRoute: initialRoute,
        labels: labels,
        onLabel: onLabel,
        materialAppFactory: materialAppFactory));
